# coding: utf-8
from typing import List, Dict, Optional, Any

from delta_trace_db.query.cause.cause import Cause
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query import Query
from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.query.nodes.query_node import QueryNode
from delta_trace_db.query.sort.abstract_sort import AbstractSort
from delta_trace_db.query.merge_query_params import MergeQueryParams


class QueryBuilder:
    def __init__(self,
                 target: str,
                 type_: EnumQueryType,
                 add_data: Optional[List[CloneableFile]] = None,
                 override_data: Optional[Dict[str, Any]] = None,
                 template: Optional[Dict[str, Any]] = None,
                 query_node: Optional[QueryNode] = None,
                 sort_obj: Optional[AbstractSort] = None,
                 offset: Optional[int] = None,
                 start_after: Optional[Dict[str, Any]] = None,
                 end_before: Optional[Dict[str, Any]] = None,
                 rename_before: Optional[str] = None,
                 rename_after: Optional[str] = None,
                 limit: Optional[int] = None,
                 return_data: bool = False,
                 must_affect_at_least_one: bool = True,
                 serial_key: Optional[str] = None,
                 reset_serial: bool = False,
                 merge_query_params: Optional[MergeQueryParams] = None,
                 cause: Optional[Cause] = None):
        """
        (en) A builder class for easily constructing queries.
        In addition to constructors for creating each query,
        there are methods for dynamically changing the paging parameters.
        For information about the parameters of this class, see the individual class methods.

        (ja) クエリを簡単に組み立てるためのビルダークラスです。
        各クエリを作成するためのコンストラクタの他に、
        ページング用のパラメータを動的に変更するためのメソッドがあります。
        本クラスのパラメータについては、個別のクラスメソッドを参照してください。
        """
        self.target = target
        self.type = type_
        self.add_data = add_data
        self.override_data = override_data
        self.template = template
        self.query_node = query_node
        self.sort_obj = sort_obj
        self.offset = offset
        self.start_after = start_after
        self.end_before = end_before
        self.rename_before = rename_before
        self.rename_after = rename_after
        self.limit = limit
        self.return_data = return_data
        self.must_affect_at_least_one = must_affect_at_least_one
        self.serial_key = serial_key
        self.reset_serial = reset_serial
        self.merge_query_params = merge_query_params
        self.cause = cause

    @classmethod
    def add(cls, target: str,
            add_data: List[CloneableFile],
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            serial_key: Optional[str] = None,
            cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Adds an item to the specified collection.
        If the specified collection does not already exist,
        it will be created automatically.

        (ja) 指定されたコレクションに要素を追加します。
        指定されたコレクションがまだ存在しない場合はコレクションが自動で作成されます。

        Parameters
        ----------
        target: str
            The collection name in DB.
        add_data: List[CloneableFile]
            Data specified when performing an add operation.
            Typically, this is assigned the list that results from calling toDict on
            a subclass of ClonableFile.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        serial_key: Optional[str]
            If not null, the add query will assign a unique serial
            number (integer value) to the specified key.
            This value is unique per collection.
            Note that only variables directly under the class can be specified as
            keys, not nested fields.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.add,
                   add_data=add_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   serial_key=serial_key,
                   cause=cause)

    @classmethod
    def update(cls, target: str,
               query_node: QueryNode,
               override_data: Dict[str, Any],
               return_data: bool = False,
               sort_obj: Optional[AbstractSort] = None,
               must_affect_at_least_one: bool = True,
               cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Overwrites the parameters of all objects in the specified collection
        that match the conditions.
        Parameters not specified for overwriting remain unchanged.

        (ja) 指定されたコレクションのうち、条件にマッチする全てのオブジェクトのパラメータを
        上書きします。上書き対象に指定していないパラメータは変化しません。

        Parameters
        ----------
        target: str
            The collection name in DB.
        query_node: QueryNode
            This is the node object used for the search.
            You can build queries by combining the various nodes defined in
            comparison_node.dart.
        override_data: Dict[str, Any]
            This is not a serialized version of the full class,
            but a dictionary containing only the parameters you want to update.
            The parameters directly below will be updated.
            For example, if the original data is {"a": 0, "b": {"c": 1}},
            and you update it by data of {"b": {"d": 2}},
            the result will be {"a": 0, "b": {"d": 2}}.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        sort_obj: Optional[AbstractSort]
            An object for sorting the return values.

            - SingleSort or MultiSort can be used.

            - Optional. If omitted, results will be returned in the order they were added to the database.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.update,
                   query_node=query_node,
                   override_data=override_data,
                   return_data=return_data,
                   sort_obj=sort_obj,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def update_one(cls, target: str,
                   query_node: QueryNode,
                   override_data: Dict[str, Any],
                   return_data: bool = False,
                   must_affect_at_least_one: bool = True,
                   cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Overwrites the parameters of one object in the specified collection
        that matches the conditions. Parameters not specified for overwriting
        remain unchanged.

        (ja) 指定されたコレクションのうち、条件にマッチする１つのオブジェクトのパラメータを
        上書きします。上書き対象に指定していないパラメータは変化しません。

        Parameters
        ----------
        target: str
            The collection name in DB.
        query_node: QueryNode
            This is the node object used for the search.
            You can build queries by combining the various nodes defined in
            comparison_node.dart.
        override_data: Dict[str, Any]
            This is not a serialized version of the full class,
            but a dictionary containing only the parameters you want to update.
            The parameters directly below will be updated.
            For example, if the original data is {"a": 0, "b": {"c": 1}},
            and you update it by data of {"b": {"d": 2}},
            the result will be {"a": 0, "b": {"d": 2}}.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.updateOne,
                   query_node=query_node,
                   override_data=override_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def delete(cls, target: str,
               query_node: QueryNode,
               return_data: bool = False,
               sort_obj: Optional[AbstractSort] = None,
               must_affect_at_least_one: bool = True,
               cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Deletes all objects in the specified collection that match
        the specified criteria.

        (ja) 指定されたコレクションのうち、条件にマッチするオブジェクトを全て削除します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        query_node: QueryNode
            This is the node object used for the search.
            You can build queries by combining the various nodes defined in
            comparison_node.dart.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        sort_obj: Optional[AbstractSort]
            An object for sorting the return values.

            - SingleSort or MultiSort can be used.

            - Optional. If omitted, results will be returned in the order they were added to the database.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.delete,
                   query_node=query_node,
                   return_data=return_data,
                   sort_obj=sort_obj,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def delete_one(cls, target: str,
                   query_node: QueryNode,
                   return_data: bool = False,
                   must_affect_at_least_one: bool = True,
                   cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Deletes only one object that matches the specified criteria from
        the specified collection.

        (ja) 指定されたコレクションのうち、条件にマッチするオブジェクトを１件だけ削除します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        query_node: QueryNode
            This is the node object used for the search.
            You can build queries by combining the various nodes defined in
            comparison_node.dart.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.deleteOne,
                   query_node=query_node,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def search(cls, target: str,
               query_node: QueryNode,
               sort_obj: Optional[AbstractSort] = None,
               offset: Optional[int] = None,
               start_after: Optional[Dict[str, Any]] = None,
               end_before: Optional[Dict[str, Any]] = None,
               limit: Optional[int] = None,
               cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Gets objects from the specified collection that match
        the specified criteria.

        (ja) 指定されたコレクションから、条件にマッチするオブジェクトを取得します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        query_node: QueryNode
            This is the node object used for the search.
            You can build queries by combining the various nodes defined in
            comparison_node.dart.
        sort_obj: Optional[AbstractSort]
            An object for sorting the return values.

            - SingleSort or MultiSort can be used.

            - Optional. If omitted, results will be returned in the order they were added to the database.
        offset: Optional[int]
            Offset for front-end paging support.
            If specified, data from the specified offset onwards will be retrieved.
        start_after: Optional[Dict[str, Any]]
            If you pass in a serialized version of a search result
            object, the search will return results from objects after that object,
            and if an offset is specified, it will be ignored.
            This does not work if there are multiple identical objects because it
            compares the object values, and is slightly slower than specifying an
            offset, but it works fine even if new objects are added during the search.
        end_before: Optional[Dict[str, Any]]
            If you pass in a serialized version of a search result
            object, the search will return results from the object before that one,
            and any offset or startAfter specified will be ignored.
            This does not work if there are multiple identical objects because it
            compares the object values, and is slightly slower than specifying an
            offset, but it works fine even if new objects are added during the search.
        limit: Optional[int]
            The maximum number of search results.

            - With offset/startAfter: returns up to [limit] items after the specified position.

            - With endBefore: returns up to [limit] items before the specified position.

            - If no offset/startAfter/endBefore is specified, the first [limit] items in addition order are returned.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.search,
                   query_node=query_node,
                   sort_obj=sort_obj,
                   offset=offset,
                   start_after=start_after,
                   end_before=end_before,
                   limit=limit,
                   cause=cause)

    @classmethod
    def search_one(cls, target: str,
                   query_node: QueryNode,
                   cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Gets objects from the specified collection that match
        the specified criteria.
        It is faster than a "search query" when searching for a single item
        because the search stops once a hit is found.

        (ja) 指定されたコレクションから、条件にマッチするオブジェクトを取得します。
        1件のヒットがあった時点で探索を打ち切るため、
        単一のアイテムを検索する場合はsearchよりも高速に動作します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        query_node: QueryNode
            This is the node object used for the search.
            You can build queries by combining the various nodes defined in
            comparison_node.dart.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.searchOne,
                   query_node=query_node,
                   cause=cause)

    @classmethod
    def get_all(cls, target: str,
                sort_obj: Optional[AbstractSort] = None,
                offset: Optional[int] = None,
                start_after: Optional[Dict[str, Any]] = None,
                end_before: Optional[Dict[str, Any]] = None,
                limit: Optional[int] = None,
                cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Gets all items in the specified collection.
        If a limit(limit, offset, startAfter, endBefore, limit) is set,
        items from the specified location and quantity will be retrieved from
        all items.

        (ja) 指定されたコレクションの全てのアイテムを取得します。
        制限(limit, offset, startAfter, endBefore, limit)をかけた場合は、
        全てのアイテムから指定の位置と量のアイテムを取得します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        sort_obj: Optional[AbstractSort]
            An object for sorting the return values.

            - SingleSort or MultiSort can be used.

            - Optional. If omitted, results will be returned in the order they were added to the database.
        offset: Optional[int]
            Offset for front-end paging support.
            If specified, data from the specified offset onwards will be retrieved.
        start_after: Optional[Dict[str, Any]]
            If you pass in a serialized version of a search result
            object, the search will return results from objects after that object,
            and if an offset is specified, it will be ignored.
            This does not work if there are multiple identical objects because it
            compares the object values, and is slightly slower than specifying an
            offset, but it works fine even if new objects are added during the search.
        end_before: Optional[Dict[str, Any]]
            If you pass in a serialized version of a search result
            object, the search will return results from the object before that one,
            and any offset or startAfter specified will be ignored.
            This does not work if there are multiple identical objects because it
            compares the object values, and is slightly slower than specifying an
            offset, but it works fine even if new objects are added during the search.
        limit: Optional[int]
            The maximum number of search results.

            - With offset/startAfter: returns up to [limit] items after the specified position.

            - With endBefore: returns up to [limit] items before the specified position.

            - If no offset/startAfter/endBefore is specified, the first [limit] items in addition order are returned.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.getAll,
                   sort_obj=sort_obj,
                   offset=offset,
                   start_after=start_after,
                   end_before=end_before,
                   limit=limit,
                   cause=cause)

    @classmethod
    def conform_to_template(cls, target: str,
                            template: Dict[str, Any],
                            must_affect_at_least_one: bool = True,
                            cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Formats the contents of the specified collection to match the
        specified template.
        Fields not present in the template will be removed,
        and fields that are only present in the template will be added
        with the template's value as their initial value.

        (ja) 指定されたコレクションの内容を、指定したテンプレートに一致するように整形します。
        テンプレートに存在しないフィールドは削除され、テンプレートにのみ存在するフィールドは、
        テンプレートの値を初期値として追加されます。

        Parameters
        ----------
        target: str
            The collection name in DB.
        template: Dict[str, Any]
            Specify this when changing the structure of the DB class.
            Fields that do not exist in the existing structure but exist in the
            template will be added with the template value as the initial value.
            Fields that do not exist in the template will be deleted.
            Usually, you pass a dictionary created by converting CloneableFile to
            Dict (call toDict).
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.conformToTemplate,
                   template=template,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def rename_field(cls, target: str,
                     rename_before: str,
                     rename_after: str,
                     return_data: bool = False,
                     must_affect_at_least_one: bool = True,
                     cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Renames a specific field in the specified collection.

        (ja) 指定されたコレクションの特定のフィールドの名前を変更します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        rename_before: str
            The old variable name when querying for a rename.
        rename_after: str
            The new name of the variable when querying for a rename.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.renameField,
                   rename_before=rename_before,
                   rename_after=rename_after,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def count(cls, target: str,
              cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Gets the number of elements in the specified collection.

        (ja) 指定されたコレクションの要素数を取得します。

        Parameters
        ----------
        target: str
            The collection name in DB.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.count,
                   cause=cause)

    @classmethod
    def clear(cls, target: str,
              must_affect_at_least_one: bool = True,
              reset_serial: bool = False,
              cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) This query empties the contents of the specified collection.

        (ja) このクエリは指定したコレクションの内容を空にします。

        Parameters
        ----------
        target: str
            The collection name in DB.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        reset_serial: bool
            If true, resets the managed serial number to 0 on
            a clear or clearAdd query.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.clear,
                   must_affect_at_least_one=must_affect_at_least_one,
                   reset_serial=reset_serial,
                   cause=cause)

    @classmethod
    def clear_add(cls, target: str,
                  add_data: List[CloneableFile],
                  return_data: bool = False,
                  must_affect_at_least_one: bool = True,
                  serial_key: Optional[str] = None,
                  reset_serial: bool = False,
                  cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Clears the specified collection and then add data.

        (ja) 指定されたコレクションをclearした後、dataをAddします。

        Parameters
        ----------
        target: str
            The collection name in DB.
        add_data: List[CloneableFile]
            Data specified when performing an add operation.
            Typically, this is assigned the list that results from calling toDict on
            a subclass of ClonableFile.
        return_data: bool
            If true, return the changed objs.
            If serialKey is set, the object will be returned with
            the serial number added.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        serial_key: Optional[str]
            If not null, the add query will assign a unique serial
            number (integer value) to the specified key.
            This value is unique per collection.
            Note that only variables directly under the class can be specified as
            keys, not nested fields.
        reset_serial: bool
            If true, resets the managed serial number to 0 on
            a clear or clearAdd query.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.clearAdd,
                   add_data=add_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   serial_key=serial_key,
                   reset_serial=reset_serial,
                   cause=cause)

    @classmethod
    def remove_collection(cls, target: str,
                          must_affect_at_least_one: bool = True,
                          cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en) Deletes the specified collection.
        This query is special because it deletes the collection itself.
        Therefore, it cannot be included as part of a transaction query.
        Additionally, any callbacks associated with the target collection will
        not be called when executed.
        This is a maintenance function for administrators who need to change
        the database structure.
        Typically, the database should be designed so that it never needs to be
        called.

        (ja) 指定されたコレクションを削除します。
        このクエリは特殊で、コレクションそのものが削除されるため
        トランザクションクエリの一部として含めることはできません。
        また、実行時には対象のコレクションに紐付いたコールバックも呼ばれません。
        これはDBの構造変更が必要な管理者のためのメンテナンス機能であり、
        通常はこれを呼び出さないでも問題ない設計にしてください。

        Parameters
        ----------
        target: str
            The collection name in DB.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(target, EnumQueryType.removeCollection,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def merge(cls, merge_query_params: MergeQueryParams, must_affect_at_least_one: bool = True,
              cause: Optional[Cause] = None) -> "QueryBuilder":
        """
        (en)
        Merges collections according to the specified template and creates
        a new collection.

        This query is special and cannot be included as part of a transaction
        query. Also, any callbacks associated with the target collection will
        not be called when it is executed.

        This is a maintenance function for administrators who need to change
        the database structure. Normally, you should design your database so
        that it does not need to be called.

        (ja)
        指定されたテンプレートに沿ってコレクションをマージし、
        新しいコレクションを作成します。

        このクエリは特殊で、トランザクションクエリの一部として
        含めることはできません。
        また、実行時には対象のコレクションに紐付いたコールバックも
        呼ばれません。

        これは DB の構造変更が必要な管理者のための
        メンテナンス機能であり、通常はこれを呼び出さないでも
        問題ない設計にしてください。

        Parameters
        ----------
        merge_query_params: MergeQueryParams
            Parameter object specifically for merge queries.
        must_affect_at_least_one: bool
            If true, the operation will be marked as
            failed if it affects 0 objects.
            If the operation is treated as a failure, the isSuccess flag of the
            returned QueryResult will be set to false.
        cause: Optional[Cause]
            Optional metadata for auditing or logging.
            Useful in high-security environments or for autonomous AI programs
            to record the reason or initiator of a query.
        """
        return cls(
            target=merge_query_params.base,
            type_=EnumQueryType.merge,
            merge_query_params=merge_query_params,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause,
        )

    def set_offset(self, new_offset: Optional[int]) -> "QueryBuilder":
        """
        (en) This method can be used if you want to change only the search position.

        (ja) 検索位置だけを変更したい場合に利用できるメソッドです。

        Parameters
        ----------
        new_offset: Optional[int]
            Offset for front-end paging support.
            If specified, data from the specified offset onwards will be retrieved.
        """
        self.offset = new_offset
        return self

    def set_start_after(self, new_start_after: Optional[Dict[str, Any]]) -> "QueryBuilder":
        """
        (en) This method can be used if you want to change only the search position.

        (ja) 検索位置だけを変更したい場合に利用できるメソッドです。

        Parameters
        ----------
        new_start_after: Optional[Dict[str, Any]]
            If you pass in a serialized version of a search result
            object, the search will return results from objects after that object,
            and if an offset is specified, it will be ignored.
            This does not work if there are multiple identical objects because it
            compares the object values, and is slightly slower than specifying an
            offset, but it works fine even if new objects are added during the search.
        """
        self.start_after = new_start_after
        return self

    def set_end_before(self, new_end_before: Optional[Dict[str, Any]]) -> "QueryBuilder":
        """
        (en) This method can be used if you want to change only the search position.

        (ja) 検索位置だけを変更したい場合に利用できるメソッドです。

        Parameters
        ----------
        new_end_before: Optional[Dict[str, Any]]
            If you pass in a serialized version of a search result
            object, the search will return results from the object before that one,
            and any offset or startAfter specified will be ignored.
            This does not work if there are multiple identical objects because it
            compares the object values, and is slightly slower than specifying an
            offset, but it works fine even if new objects are added during the search.
        """
        self.end_before = new_end_before
        return self

    def set_limit(self, new_limit: Optional[int]) -> "QueryBuilder":
        """
        (en) This method can be used if you want to change only the limit.

        (ja) limitだけを変更したい場合に利用できるメソッドです。

        Parameters
        ----------
        new_limit: Optional[int]
            The maximum number of search results.

            - With offset/startAfter: returns up to [limit] items after the specified position.

            - With endBefore: returns up to [limit] items before the specified position.

            - If no offset/startAfter/endBefore is specified, the first [limit] items in addition order are returned.
        """
        self.limit = new_limit
        return self

    def build(self) -> Query:
        """
        (en) Commit the content and convert it into a query object.

        (ja) 内容を確定してクエリーオブジェクトに変換します。
        """
        m_data: Optional[List[Dict[str, Any]]] = None
        if self.add_data is not None:
            m_data = [i.to_dict() for i in self.add_data]

        return Query(
            target=self.target,
            type_=self.type,
            add_data=m_data,
            override_data=self.override_data,
            template=self.template,
            query_node=self.query_node,
            sort_obj=self.sort_obj,
            offset=self.offset,
            start_after=self.start_after,
            end_before=self.end_before,
            rename_before=self.rename_before,
            rename_after=self.rename_after,
            limit=self.limit,
            return_data=self.return_data,
            must_affect_at_least_one=self.must_affect_at_least_one,
            serial_key=self.serial_key,
            reset_serial=self.reset_serial,
            merge_query_params=self.merge_query_params,
            cause=self.cause
        )



# coding: utf-8
from threading import RLock
from typing import Any, Dict, List, Callable, Optional, override

from file_state_manager.cloneable_file import CloneableFile

from delta_trace_db.query.raw_query_builder import RawQueryBuilder
from delta_trace_db.db.delta_trace_db_collection import Collection
from delta_trace_db.dsl.util_dsl_evaluator import UtilDslEvaluator
from delta_trace_db.query.cause.permission import Permission
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query import Query
from delta_trace_db.query.query_execution_result import QueryExecutionResult
from delta_trace_db.query.query_result import QueryResult
from delta_trace_db.query.transaction_query import TransactionQuery
from delta_trace_db.query.transaction_query_result import TransactionQueryResult
from delta_trace_db.query.util_query import UtilQuery
import logging

_logger = logging.getLogger(__name__)


class DeltaTraceDatabase(CloneableFile):
    class_name = "DeltaTraceDatabase"
    version = "16"

    def __init__(self):
        """
        (en) It is an in-memory database that takes into consideration the
        safety of various operations.
        It was created with the assumption that in addition to humans,
        AI will also be the main users.

        (ja) 様々な操作の安全性を考慮したインメモリデータベースです。
        人間以外で、AIも主な利用者であると想定して作成しています。
        """
        super().__init__()
        self._collections: Dict[str, Collection] = {}
        self._lock = RLock()  # execute_query / execute_transaction_query 共通

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "DeltaTraceDatabase":
        """
        (en) Restore this object from the dictionary.
        Note that src is used as is, not copied.

        (ja) このオブジェクトを辞書から復元します。
        srcはコピーされずにそのまま利用されることに注意してください。

        Parameters
        ----------
        src : Dict[str, Any]
            A dictionary made with toDict of this class.

        Raises
        ------
        ValueError
            Throws on ValueError if the src is invalid format.
        """
        instance = cls()
        instance._collections = cls._parse_collections(src)
        return instance

    @staticmethod
    def _parse_collections(src: Dict[str, Any]) -> Dict[str, Collection]:
        """
        (en) Restoring database data from JSON.

        (ja) データベースのJSONからの復元処理。

        Parameters
        ----------
        src : Dict[str, Any]
            A dictionary made with toDict of this class.

        Raises
        ------
        ValueError
            Throws on ValueError if the src is invalid format.
        """
        cols = src.get("collections")
        if not isinstance(cols, dict):
            raise ValueError("Invalid format: 'collections' should be a dict")
        result: Dict[str, Collection] = {}
        for key, value in cols.items():
            if not isinstance(value, dict):
                raise ValueError("Invalid format: target is not a dict")
            result[key] = Collection.from_dict(value)
        return result

    def collection(self, name: str) -> Collection:
        """
        (en) If the specified collection exists, it will be retrieved.
        If it does not exist, a new one will be created and retrieved.
        Normally you should not call this directly, but rather operate via queries.

        (ja) 指定のコレクションが存在すればそれを取得し、
        存在しなければ新しく作成して取得します。
        通常は直接これを呼び出さず、クエリ経由で操作してください。

        Parameters
        ----------
        name : str
            The collection name.
        """
        with self._lock:
            if name in self._collections:
                return self._collections[name]
            col = Collection()
            self._collections[name] = col
            return col

    def find_collection(self, name: str) -> Optional[Collection]:
        """
        (en) Find the specified collection.
        Returns it if it exists, otherwise returns None.

        (ja) 指定のコレクションを検索します。
        存在すれば返し、存在しなければ None を返します。

        Parameters
        ----------
        name : str
            The collection name.
        """
        with self._lock:
            return self._collections.get(name)

    def remove_collection(self, name: str) -> None:
        """
        (en) Deletes the specified collection.
        If a collection with the specified name does not exist, this does nothing.

        (ja) 指定のコレクションを削除します。
        指定の名前のコレクションが存在しなかった場合は何もしません。

        Parameters
        ----------
        name : str
            The collection name.

        Raises
        ------
        ValueError
            例外が発生する条件があれば説明。
        """
        with self._lock:
            self._collections.pop(name, None)

    def collection_to_dict(self, name: str) -> Optional[Dict[str, Any]]:
        """
        (en) Saves individual collections as dictionaries.
        For example, you can use this if you want to store a specific collection
        in an encrypted format.
        If you specify a collection that does not exist, null is returned.

        (ja) 個別のコレクションを辞書として保存します。
        特定のコレクション単位で暗号化して保存したいような場合に利用できます。
        存在しないコレクションを指定した場合はnullが返されます。

        Parameters
        ----------
        name : str
            The collection name.
        """
        with self._lock:
            collection = self._collections.get(name)
            return collection.to_dict() if collection is not None else None

    def collection_from_dict(self, name: str, src: Dict[str, Any]) -> Collection:
        """
        (en) Restores a specific collection from a dictionary, re-registers it,
        and retrieves it.
        If a collection with the same name already exists, it will be overwritten.
        This is typically used to restore data saved with collectionToDict.


        (ja) 特定のコレクションを辞書から復元して再登録し、取得します。
        既存の同名のコレクションが既にある場合は上書きされます。
        通常は、collectionToDictで保存したデータを復元する際に使用します。

        Parameters
        ----------
        name : str
            The collection name.
        src : Dict[str, Any]
            A dictionary made with collectionToDict of this class.

        Raises
        ------
        ValueError
            Throws on ValueError if the src is invalid format.
        """
        with self._lock:
            col = Collection.from_dict(src)
            self._collections[name] = col
            return col

    def collection_from_dict_keep_listener(self, name: str, src: Dict[str, Any]) -> Collection:
        """
        (en) Restores a specific collection from a dictionary, re-registers it,
        and retrieves it.
        If a collection with the same name already exists, it will be overwritten.
        This is typically used to restore data saved with collection_to_dict.
        This method preserves existing listeners when overwriting the specified
        collection.

        (ja) 特定のコレクションを辞書から復元して再登録し、取得します。
        既存の同名のコレクションが既にある場合は上書きされます。
        通常は、collection_to_dictで保存したデータを復元する際に使用します。
        このメソッドでは、指定されたコレクションの上書き時、既存のリスナが維持されます。

        Parameters
        ----------
        name : str
            The collection name.
        src : Dict[str, Any]
            A dictionary made with collectionToDict of this class.

        Raises
        ------
        ValueError
            Throws on ValueError if the src is invalid format.
        """
        with self._lock:
            col = Collection.from_dict(src)
            listeners_buf = None
            named_listeners_buf = None
            if name in self._collections:
                listeners_buf = self._collections[name].listeners
                named_listeners_buf = self._collections[name].named_listeners
            self._collections[name] = col
            if listeners_buf is not None:
                col.listeners = listeners_buf
            if named_listeners_buf is not None:
                col.named_listeners = named_listeners_buf
            return col

    @override
    def clone(self) -> "DeltaTraceDatabase":
        with self._lock:
            return DeltaTraceDatabase.from_dict(self.to_dict())

    @property
    def raw(self) -> Dict[str, Collection]:
        """
        (en) Returns the stored contents as a reference.
        Be careful as it is dangerous to edit it directly.

        (ja) 保持している内容を参照として返します。
        直接編集すると危険なため注意してください。
        """
        return self._collections

    @override
    def to_dict(self) -> Dict[str, Any]:
        with self._lock:
            return {
                "className": self.class_name,
                "version": self.version,
                "collections": {k: v.to_dict() for k, v in self._collections.items()},
            }

    def add_listener(self, target: str, cb: Callable[[], None], name: Optional[str] = None):
        """
        (en) This is a callback setting function that can be used when linking
        the UI and DB.
        The callback set here will be called when the contents of the [target]
        collection are changed.
        In other words, if you register it, you will be able to update the screen,
        etc. when the contents of the DB change.
        Normally you would register it in initState and then use removeListener
        to remove it when disposing.
        If you use this on the server side, it may be a good idea to set up a
        function that writes the backup to storage.
        Please note that notifications will not be restored even if the DB is
        deserialized. You will need to set them every time.

        (ja) UIとDBを連携する際に利用できる、コールバックの設定関数です。
        ここで設定したコールバックは、[target]のコレクションの内容が変更されると呼び出されます。
        つまり、登録しておくとDBの内容変更時に画面更新等ができるようになります。
        通常はinitStateで登録し、dispose時にremoveListenerを使って解除してください。
        これをサーバー側で使用する場合は、バックアップをストレージに書き込む関数などを設定
        するのも良いかもしれません。
        なお、通知に関してはDBをデシリアライズしても復元されません。毎回設定する必要があります。

        Parameters
        ----------
        target : str
            The target collection name.
        cb : Callable[[], None]
            The function to execute when the DB is changed.
        name : Optional[str]
            If you set a non-null value, a listener will be registered with that name.
            Setting a name is useful if you want to be more precise about registration and release.
        """
        with self._lock:
            self.collection(target).add_listener(cb, name=name)

    def remove_listener(self, target: str, cb: Callable[[], None], name: Optional[str] = None):
        """
        (en) This function is used to cancel the set callback.
        Call it in the UI using dispose etc.

        (ja) 設定したコールバックを解除するための関数です。
        UIではdisposeなどで呼び出します。

        Parameters
        ----------
        target : str
            The target collection name.
        cb : Callable[[], None]
            The function for which you want to cancel the notification.
        name : Optional[str]
            If you registered with a name when you added Listener, you must unregister with the same name.
        """
        with self._lock:
            self.collection(target).remove_listener(cb, name=name)

    def execute_query_object(self, query: Any,
                             collection_permissions: Optional[Dict[str, Permission]] = None) -> QueryExecutionResult:
        """
        (en) Executes a query of any type.
        This function can execute a regular query, a transactional query,
        or a Dict of any of these.
        Server side, verify that the call is legitimate
        (e.g. by checking the JWT and/or the caller's user permissions)
        before making this call.

        (ja) 型を問わずにクエリを実行します。
        この関数は、通常のクエリ、トランザクションクエリ、
        またはそれらをDictにしたもののいずれでも実行できます。
        サーバーサイドでは、この呼び出しの前に正規の呼び出しであるかどうかの
        検証(JWTのチェックや呼び出し元ユーザーの権限のチェック)を行ってください。

        Parameters
        ----------
        query : Any
            Query, TransactionQuery, or Dict.
        collection_permissions: Optional[Dict[str, Permission]]
            Collection level operation permissions for the executing user. This is an optional argument for the server,
            the key is the target collection name. Use null on the frontend, if this is null then everything is allowed.

        Raises
        ------
        ValueError
            Throws on ValueError if the query is unsupported type.
        """
        with self._lock:  # 排他制御
            if isinstance(query, Query):
                return self.execute_query(query, collection_permissions=collection_permissions)
            elif isinstance(query, TransactionQuery):
                return self.execute_transaction_query(query, collection_permissions=collection_permissions)
            elif isinstance(query, dict):
                if query.get("className") == "Query":
                    return self.execute_query(Query.from_dict(query), collection_permissions=collection_permissions)
                elif query.get("className") == "TransactionQuery":
                    return self.execute_transaction_query(TransactionQuery.from_dict(query),
                                                          collection_permissions=collection_permissions)
                else:
                    raise ValueError("Unsupported query class")
            else:
                raise ValueError("Unsupported query type")

    def execute_query(self, q: Query, collection_permissions: Optional[Dict[str, Permission]] = None) -> QueryResult:
        """
        (en) Execute the query.
        Server side, verify that the call is legitimate
        (e.g. by checking the JWT and/or the caller's user permissions)
        before making this call.

        (ja) クエリを実行します。
        サーバーサイドでは、この呼び出しの前に正規の呼び出しであるかどうかの
        検証(JWTのチェックや呼び出し元ユーザーの権限のチェック)を行ってください。

        Parameters
        ----------
        q : Query
            The query.
        collection_permissions: Optional[Dict[str, Permission]]
            Collection level operation permissions for the executing user. This is an optional argument for the server,
            the key is the target collection name. Use null on the frontend, if this is null then everything is allowed.
        """
        with self._lock:  # 単体クエリもここで排他
            try:
                # パーミッションのチェック
                if not UtilQuery.check_permissions(q=q, collection_permissions=collection_permissions):
                    return QueryResult(
                        is_success=False,
                        target=q.target,
                        type_=q.type,
                        result=[],
                        db_length=-1,
                        update_count=0,
                        hit_count=0,
                        error_message="Operation not permitted."
                    )
                is_exist_col = q.target in self._collections
                col = self.collection(q.target)
                match q.type:
                    case EnumQueryType.add:
                        r = col.add_all(q)
                    case EnumQueryType.update:
                        r = col.update(q, is_single_target=False)
                    case EnumQueryType.updateOne:
                        r = col.update(q, is_single_target=True)
                    case EnumQueryType.delete:
                        r = col.delete(q)
                    case EnumQueryType.deleteOne:
                        r = col.delete_one(q)
                    case EnumQueryType.search:
                        r = col.search(q)
                    case EnumQueryType.searchOne:
                        r = col.search_one(q)
                    case EnumQueryType.getAll:
                        r = col.get_all(q)
                    case EnumQueryType.conformToTemplate:
                        r = col.conform_to_template(q)
                    case EnumQueryType.renameField:
                        r = col.rename_field(q)
                    case EnumQueryType.count:
                        r = col.count(q)
                    case EnumQueryType.clear:
                        r = col.clear(q)
                    case EnumQueryType.clearAdd:
                        r = col.clear_add(q)
                    case EnumQueryType.removeCollection:
                        if is_exist_col:
                            r = QueryResult(
                                is_success=True,
                                target=q.target,
                                type_=q.type,
                                result=[],
                                db_length=0,
                                update_count=1,
                                hit_count=0,
                                error_message=None)
                        else:
                            r = QueryResult(
                                is_success=True,
                                target=q.target,
                                type_=q.type,
                                result=[],
                                db_length=0,
                                update_count=0,
                                hit_count=0,
                                error_message=None)
                        self.remove_collection(name=q.target)
                    case EnumQueryType.merge:
                        r = self._execute_merge_query(q=q)
                # must_affect_at_least_oneの判定。
                if q.type in (
                        EnumQueryType.add,
                        EnumQueryType.update,
                        EnumQueryType.updateOne,
                        EnumQueryType.delete,
                        EnumQueryType.deleteOne,
                        EnumQueryType.conformToTemplate,
                        EnumQueryType.renameField,
                        EnumQueryType.clear,
                        EnumQueryType.clearAdd,
                        EnumQueryType.removeCollection,
                        EnumQueryType.merge
                ):
                    if q.must_affect_at_least_one and r.update_count == 0 and r.is_success:
                        return QueryResult(
                            is_success=False,
                            target=q.target,
                            type_=q.type,
                            result=[],
                            db_length=len(col.raw),
                            update_count=0,
                            hit_count=r.hit_count,
                            error_message="No data matched the condition (mustAffectAtLeastOne=True)"
                        )
                return r
            except ValueError:
                # ここでは安全なメッセージのみを外部に返す
                _logger.error("execute_query ArgumentError", exc_info=True)
                return QueryResult(
                    is_success=False,
                    target=q.target,
                    type_=q.type,
                    result=[],
                    db_length=-1,
                    update_count=0,
                    hit_count=0,
                    error_message="execute_query ArgumentError"
                )
            except Exception:
                _logger.error("execute_query Unexpected Error", exc_info=True)
                return QueryResult(
                    is_success=False,
                    target=q.target,
                    type_=q.type,
                    result=[],
                    db_length=-1,
                    update_count=0,
                    hit_count=0,
                    error_message="execute_query Unexpected Error",
                )

    def execute_transaction_query(self, q: TransactionQuery,
                                  collection_permissions: Optional[
                                      Dict[str, Permission]] = None) -> TransactionQueryResult:
        """
        (en) Execute the transaction query.
        Server side, verify that the call is legitimate
        (e.g. by checking the JWT and/or the caller's user permissions)
        before making this call.
        During a transaction, after all operations are completed successfully,
        if there is a listener callback for each collection, it will be invoked.
        If there is a failure, nothing will be done.
        removeCollection and merge queries cannot be executed and will return an
        error message if they are included in a transaction.

        (ja) トランザクションクエリを実行します。
        サーバーサイドでは、この呼び出しの前に正規の呼び出しであるかどうかの
        検証(JWTのチェックや呼び出し元ユーザーの権限のチェック)を行ってください。
        トランザクション時は、全ての処理が正常に完了後、各コレクションに
        リスナーのコールバックがあれば起動し、失敗の場合はなにもしません。
        removeCollection及びmergeクエリは実行できず、
        トランザクションに含まれる場合はエラーメッセージが返されます。

        Parameters
        ----------
        q : Query
            The query.
        collection_permissions: Optional[Dict[str, Permission]]
            Collection level operation permissions for the executing user. This is an optional argument for the server,
            the key is the target collection name. Use null on the frontend, if this is null then everything is allowed.
        """
        with self._lock:  # トランザクション全体で排他
            # 許可されていないクエリが混ざっていないか調査し、混ざっていたら失敗にする。
            for i in q.queries:
                if i.type == EnumQueryType.removeCollection or i.type == EnumQueryType.merge:
                    return TransactionQueryResult(is_success=False, results=[],
                                                  error_message="The query contains a type that is not permitted to be executed within a transaction.")
            # トランザクション付き処理を開始。
            results: List[QueryResult] = []
            try:
                buff: Dict[str, Dict[str, Any]] = {}
                non_exist_targets: set[str] = set()
                for i in q.queries:
                    if i.target in buff:
                        continue
                    else:
                        t_collection: Optional[dict[str, Any]] = self.collection_to_dict(i.target)
                        if t_collection is not None:
                            buff[i.target] = t_collection
                            # コレクションをトランザクションモードに変更する。
                            self.collection(i.target).change_transaction_mode(True)
                        else:
                            non_exist_targets.add(i.target)
                try:
                    for i in q.queries:
                        results.append(self.execute_query(i, collection_permissions=collection_permissions))
                except Exception:
                    _logger.error("execute_transaction_query: Transaction failed", exc_info=True)
                    return self._rollback_collections(buff=buff, non_exist_targets=non_exist_targets)

                # rollback if any query failed
                if any(not r.is_success for r in results):
                    return self._rollback_collections(buff=buff, non_exist_targets=non_exist_targets)

                # commit: notify listeners
                for key in buff.keys():
                    col = self.collection(key)
                    need_callback = col.run_notify_listeners_in_transaction
                    col.change_transaction_mode(False)
                    if need_callback:
                        col.notify_listeners()
                return TransactionQueryResult(is_success=True, results=results)
            except Exception:
                _logger.error("execute_transaction_query: Transaction failed", exc_info=True)
                return TransactionQueryResult(is_success=False, results=[], error_message="Unexpected Error")

    def _rollback_collections(self,
                              buff: dict[str, dict[str, Any]],
                              non_exist_targets: set[str],
                              ) -> TransactionQueryResult:
        """
        (en) Rollback db.

        (ja) DBをロールバックします。

        Parameters
        ----------
        buff: dict[str, dict[str, Any]]
            The collection buffer that needs to be undone.
        non_exist_targets: set[str]
            A list of collections that did not exist before the operation.
        """
        # DBの変更を元に戻す。
        for key in buff.keys():
            self.collection_from_dict_keep_listener(key, buff[key])
            # 念のため確実に false にする。
            self.collection(key).change_transaction_mode(False)
        # 操作前に存在しなかったコレクションは削除する。
        for key in non_exist_targets:
            self.remove_collection(key)
        _logger.error("execute_transaction_query: Transaction failed", exc_info=True)
        return TransactionQueryResult(
            is_success=False,
            results=[],
            error_message="Transaction failed",
        )

    def _execute_merge_query(self, q: Query) -> QueryResult:
        """
        (en) Run merge query.

        (ja) マージクエリを実行します。

        Parameters
        ----------
        q: Query
            The query.
        """
        mqp = q.merge_query_params
        if mqp is None:
            return QueryResult(
                is_success=False,
                target=q.target,
                type_=q.type,
                result=[],
                db_length=0,
                update_count=0,
                hit_count=0,
                error_message="Argument error",
            )
        # 捜査対象コレクションの存在チェック
        if self.find_collection(mqp.base) is None:
            return QueryResult(
                is_success=False,
                target=q.target,
                type_=q.type,
                result=[],
                db_length=0,
                update_count=0,
                hit_count=0,
                error_message="Base collection does not exist.",
            )
        if mqp.source:
            for col_name in mqp.source:
                if self.find_collection(col_name) is None:
                    return QueryResult(
                        is_success=False,
                        target=q.target,
                        type_=q.type,
                        result=[],
                        db_length=0,
                        update_count=0,
                        hit_count=0,
                        error_message="Source collection does not exist.",
                    )
        if mqp.serial_base is not None:
            if self.find_collection(mqp.serial_base) is None:
                return QueryResult(
                    is_success=False,
                    target=q.target,
                    type_=q.type,
                    result=[],
                    db_length=0,
                    update_count=0,
                    hit_count=0,
                    error_message="Serial base collection does not exist.",
                )
        # フラグの設定がおかしい場合はエラー
        if len(mqp.source) != len(mqp.source_keys):
            return QueryResult(
                is_success=False,
                target=q.target,
                type_=q.type,
                result=[],
                db_length=0,
                update_count=0,
                hit_count=0,
                error_message="The relationKey or relationKeys setting is invalid.",
            )
        # 既に出力先コレクションが存在するならエラー
        if self.find_collection(mqp.output) is not None:
            return QueryResult(
                is_success=False,
                target=q.target,
                type_=q.type,
                result=[],
                db_length=0,
                update_count=0,
                hit_count=0,
                error_message="The output collection already exists.",
            )
        try:
            # DSLを解釈しながら新しいデータを生成
            new_data = []
            # マージ元データ取得
            base_collection = self.find_collection(mqp.base).raw
            source_collections = []
            for source_name in mqp.source:
                source_collections.append(
                    self.find_collection(source_name).raw
                )
            # テンプレート構造に沿ってDSLを実行
            for base_item in base_collection:
                matched_sources = UtilDslEvaluator.resolve_source_items(
                    base_item,
                    source_collections,
                    mqp.relation_key,
                    mqp.source_keys,
                )
                new_row = UtilDslEvaluator.run(
                    mqp.dsl_tmp,
                    base_item,
                    matched_sources,
                )
                new_data.append(new_row)
            # 新しいコレクションとして追加
            if mqp.serial_base is not None:
                # シリアルナンバーを引き継ぐ
                self._collections[mqp.output] = Collection.from_data(
                    new_data,
                    self.find_collection(mqp.serial_base).get_serial_num(),
                )
            else:
                # serial_key 依存でシリアル追加
                added_result = (
                    self.collection(mqp.output)
                    .add_all(
                        RawQueryBuilder.add(
                            target=mqp.output,
                            raw_add_data=new_data,
                            serial_key=mqp.serial_key,
                        ).build()
                    )
                )
                if not added_result.is_success:
                    self.remove_collection(mqp.output)
                    return QueryResult(
                        is_success=False,
                        target=q.target,
                        type_=q.type,
                        result=[],
                        db_length=0,
                        update_count=0,
                        hit_count=0,
                        error_message=added_result.error_message,
                    )
            result_col = self.find_collection(mqp.output)
            length = result_col.length if result_col is not None else 0
            return QueryResult(
                is_success=True,
                target=q.target,
                type_=q.type,
                result=[],
                db_length=length,
                update_count=length,
                hit_count=0,
            )
        except Exception:
            # 途中で失敗した場合はロールバック
            if self.find_collection(mqp.output) is not None:
                self.remove_collection(mqp.output)
            _logger.error("execute_merge_query failed", exc_info=True)
            return QueryResult(
                is_success=False,
                target=q.target,
                type_=q.type,
                result=[],
                db_length=0,
                update_count=0,
                hit_count=0,
                error_message="Merge failed",
            )



# coding: utf-8
from typing import Dict, Any

from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.query_builder import QueryBuilder
from delta_trace_db.query.sort.single_sort import SingleSort


class User(CloneableFile):
    def __init__(self, age: int):
        super().__init__()
        self.age = age

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> 'User':
        return cls(age=src['age'])

    def to_dict(self) -> Dict[str, Any]:
        return {'age': self.age}

    def clone(self) -> 'User':
        return User.from_dict(self.to_dict())


def test_get_all():
    records_count = 100
    db = DeltaTraceDatabase()
    users = [User(age=i) for i in range(records_count)]

    # データ追加
    q1 = QueryBuilder.add(target='users', add_data=users).build()
    r1 = db.execute_query(q1)
    assert r1.is_success
    assert r1.db_length == records_count

    # get_all
    q2 = QueryBuilder.get_all(target='users').build()
    r2 = db.execute_query(q2)
    assert len(r2.result) == records_count

    # limit
    q3 = QueryBuilder.get_all(target='users', limit=10).build()
    r3 = db.execute_query(q3)
    assert len(r3.result) == 10

    # offset
    q4 = QueryBuilder.get_all(target='users', limit=10, offset=10).build()
    r4 = db.execute_query(q4)
    assert len(r4.result) == 10
    assert r4.result[-1]["age"] == 19

    # start_after
    q5 = QueryBuilder.get_all(target='users', limit=10, start_after=User(age=9).to_dict()).build()
    r5 = db.execute_query(q5)
    assert len(r5.result) == 10
    assert r5.result[-1]["age"] == 19

    # end_before
    q6 = QueryBuilder.get_all(target='users', limit=10, end_before=User(age=10).to_dict()).build()
    r6 = db.execute_query(q6)
    assert len(r6.result) == 10
    assert r6.result[-1]["age"] == 9

    # offset + sort
    q7 = QueryBuilder.get_all(target='users', limit=10, offset=20,
                              sort_obj=SingleSort(field="age", reversed_=True)).build()
    r7 = db.execute_query(q7)
    assert len(r7.result) == 10
    assert r7.result[-1]["age"] == 70



from typing import Any, Dict, Optional, List, override
from datetime import datetime, timezone, timedelta

from file_state_manager import CloneableFile
from file_state_manager.util_object_hash import UtilObjectHash

from delta_trace_db.db.util_copy import UtilCopy
from delta_trace_db.query.cause.enum_actor_type import EnumActorType
from delta_trace_db.query.cause.permission import Permission


# 深いコレクション比較（Dart の DeepCollectionEquality 相当）
def deep_collection_equals(a: Any, b: Any) -> bool:
    """
    (en) Deep collection comparison function (equivalent to Dart's DeepCollectionEquality)

    (ja) 深いコレクション比較用関数（Dart の DeepCollectionEquality 相当）

    Parameters
    ----------
    a: Any
        Comparison object A.
    b: Any
        Comparison object B.
    """
    if isinstance(a, dict) and isinstance(b, dict):
        if a.keys() != b.keys():
            return False
        return all(deep_collection_equals(a[k], b[k]) for k in a)
    elif isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
        if len(a) != len(b):
            return False
        return all(deep_collection_equals(x, y) for x, y in zip(a, b))
    else:
        return a == b


def _parse_dt(src: Dict[str, Any], key: str) -> Optional[datetime]:
    v = src.get(key)
    if v is None:
        return None
    try:
        # naive datetime なら UTC として扱う
        dt = datetime.fromisoformat(v)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        else:
            dt = dt.astimezone(timezone.utc)
        return dt
    except Exception:
        return None


def _dt_to_utc_iso(dt: Optional[datetime]) -> Optional[str]:
    if dt is None:
        return None
    # naive datetime なら UTC として扱う
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    else:
        dt = dt.astimezone(timezone.utc)
    return dt.isoformat()


def _to_utc(dt: Optional[datetime]) -> Optional[datetime]:
    if dt is None:
        return None
    if dt.tzinfo is None:
        # naive datetime を UTC と見なす
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


class Actor(CloneableFile):
    className = "Actor"
    version = "6"

    def __init__(self, actor_type: EnumActorType, actor_id: str,
                 collection_permissions: Optional[Dict[str, Permission]] = None,
                 context: Optional[Dict[str, Any]] = None, name: Optional[str] = None,
                 email: Optional[str] = None,
                 created_at: Optional[datetime] = None,
                 updated_at: Optional[datetime] = None,
                 last_access: Optional[datetime] = None,
                 last_access_day: Optional[datetime] = None,
                 operation_in_day: Optional[int] = None,
                 device_ids: Optional[List[str]] = None, ):
        """
        (en) This class defines the information of the person who
        requested the database operation.

        (ja) データベースの操作をリクエストした者の情報を定義するクラスです。

        Parameters
        ----------
        actor_type: EnumActorType
            The actor type. Choose from ``human``, ``ai``, or ``system``.
        actor_id: str
            The serial id (user id) of the actor.
        collection_permissions: Optional[Dict[str, Permission]]
            Collection-level permissions that relate only to database operations.
            The key is the collection name.
        context: Optional[Dict[str, Any]]
            Additional metadata related to this actor.
        name: Optional[str]
            The actor's display name.
        email: Optional[str]
            The actor's email address.
        created_at: Optional[datetime]
            The creation timestamp (UTC) of this actor. If null, set to now (UTC).
        updated_at : Optional[datetime]
            The last update timestamp (UTC) of this actor.
            If None, set to `created_at`.
            This parameter should only be manually overridden when the values
            of `name`, `email`, `context`, or `device_ids` have been changed.
        last_access : Optional[datetime]
            The timestamp (UTC) of the last database access by this actor.
            It will be automatically updated when calling `update_access()`.
        last_access_day : Optional[datetime]
            Day-based timestamp (UTC) used to track daily operation counts.
            Automatically updated when calling `update_access()`.
        operation_in_day : int, default=0
            The number of operations performed since `last_access_day`.
            After creating an Actor instance manually,
            call `update_access()` to initialize the daily operation state.
        device_ids : Optional[List[str]]
            A list of device IDs associated with this actor.
            Used to identify devices used by the same user.
        """
        super().__init__()
        self.actor_type = actor_type
        self.actor_id = actor_id
        self.collection_permissions = collection_permissions
        self.context = context
        # added v6
        self.name = name
        self.email = email
        now = datetime.now(timezone.utc)
        self.created_at = _to_utc(created_at or now)
        self.updated_at = _to_utc(updated_at or self.created_at)
        self.last_access = _to_utc(last_access)
        self.last_access_day = _to_utc(last_access_day)
        self.operation_in_day = operation_in_day
        self.device_ids = device_ids

    def update_access(self, now: datetime, reset_hour: int = 5):
        """
        (en) Updates the access counter and last access date and time.
        Please note that updatedAt will not be updated.
        After creating an Actor instance, call this method to record the
        first access and initialize daily operation count.

        (ja) アクセスカウンタや最終アクセス日時を更新します。
        なお、updatedAtは更新されないことに注意してください。
        Actorインスタンスを作成した後、最初のアクセスを記録し
        日次操作カウントを初期化するために、このメソッドを呼んでください。

        Parameters
        ----------
        now: datetime
            Current time. If it is not UTC, it will be automatically converted to UTC.
        reset_hour: int
            Specifies the time in UTC based on which the date and time count is updated.

        Raises
        ------
        ValueError
            if the `reset_hour` is invalid value (`reset_hour` < 0 || `reset_hour` > 23) .
        """
        if reset_hour < 0 or reset_hour > 23:
            raise ValueError("reset_hour must be between 0 and 23")
        utc_now = _to_utc(now)
        today_reset = utc_now.replace(hour=reset_hour, minute=0, second=0, microsecond=0)
        if utc_now < today_reset:
            reset_base = today_reset - timedelta(days=1)
        else:
            reset_base = today_reset
        need_reset = (
                self.last_access_day is None
                or self.last_access_day < reset_base
        )
        if need_reset:
            self.operation_in_day = 1
            self.last_access_day = reset_base
        else:
            self.operation_in_day = (self.operation_in_day or 0) + 1

        self.last_access = utc_now

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "Actor":
        m_collection_permissions: Optional[Dict[str, Dict[str, Any]]] = None
        if src.get("collectionPermissions") is not None:
            m_collection_permissions = {
                k: v for k, v in src["collectionPermissions"].items()
            }

        collection_permissions: Optional[Dict[str, Permission]] = None
        if m_collection_permissions is not None:
            collection_permissions = {
                key: Permission.from_dict(value)
                for key, value in m_collection_permissions.items()
            }

        return cls(
            actor_type=EnumActorType[src["type"]],
            actor_id=src["id"],
            collection_permissions=collection_permissions,
            context=src.get("context"),
            name=src.get("name"),
            email=src.get("email"),
            created_at=_parse_dt(src, "createdAt"),
            updated_at=_parse_dt(src, "updatedAt"),
            last_access=_parse_dt(src, "lastAccess"),
            last_access_day=_parse_dt(src, "lastAccessDay"),
            operation_in_day=src.get("operationInDay"),
            device_ids=src.get("deviceIds"),
        )

    @override
    def clone(self) -> "Actor":
        return Actor.from_dict(self.to_dict())

    @override
    def to_dict(self) -> Dict[str, Any]:
        m_collection_permissions: Optional[Dict[str, Dict[str, Any]]] = None
        if self.collection_permissions is not None:
            m_collection_permissions = {
                key: value.to_dict()
                for key, value in self.collection_permissions.items()
            }

        return {
            "className": self.className,
            "version": self.version,
            "type": self.actor_type.name,
            "id": self.actor_id,
            "collectionPermissions": m_collection_permissions,
            "context": UtilCopy.jsonable_deep_copy(self.context),
            "name": self.name,
            "email": self.email,
            "createdAt": _dt_to_utc_iso(self.created_at),
            "updatedAt": _dt_to_utc_iso(self.updated_at),
            "lastAccess": _dt_to_utc_iso(self.last_access),
            "lastAccessDay": _dt_to_utc_iso(self.last_access_day),
            "operationInDay": self.operation_in_day,
            "deviceIds": UtilCopy.jsonable_deep_copy(self.device_ids),
        }

    def __eq__(self, other: object) -> bool:
        # createdAt, updatedAt, lastAccess, lastAccessDay, operationInDayは
        # ユーザーの同一性には影響しないため計算対象外。
        if not isinstance(other, Actor):
            return False
        return (
                self.actor_type == other.actor_type
                and self.actor_id == other.actor_id
                and deep_collection_equals(self.collection_permissions, other.collection_permissions)
                and deep_collection_equals(self.context, other.context)
                and self.name == other.name
                and self.email == other.email
                and deep_collection_equals(self.device_ids, other.device_ids)
        )

    def __hash__(self) -> int:
        # createdAt, updatedAt, lastAccess, lastAccessDay, operationInDayは
        # ユーザーの同一性には影響しないため計算対象外。
        return hash((
            self.actor_type,
            self.actor_id,
            UtilObjectHash.calc_map(self.collection_permissions) if self.collection_permissions else 0,
            UtilObjectHash.calc_map(self.context) if self.context else 0,
            self.name or "",
            self.email or "",
            UtilObjectHash.calc_list(self.device_ids) if self.device_ids else 0,
        ))



import importlib
import pytest

# delta_trace_db パッケージのトップレベル __all__
from delta_trace_db import __all__ as top_level_all

@pytest.mark.parametrize("name", top_level_all)
def test_top_level_imports(name):
    """
    __init__.py で定義されたトップレベル __all__ に含まれる
    名前が実際に import できるか確認する
    """
    module = importlib.import_module("delta_trace_db")
    assert hasattr(module, name), f"'{name}' is not accessible at top level"



# coding: utf-8
from abc import ABC, abstractmethod
from typing import Any, Dict
from delta_trace_db.query.nodes.enum_node_type import EnumNodeType


class QueryNode(ABC):

    @abstractmethod
    def evaluate(self, data: Dict[str, Any]) -> bool:
        """
        (en) Returns true if the object matches the calculation.

        (ja) 計算と一致するオブジェクトだった場合はtrueを返します。
        """
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """
        (en) Convert the object to a dictionary.
        The returned dictionary can only contain primitive types, null, lists
        or dicts with only primitive elements.
        If you want to include other classes,
        the target class should inherit from this class and chain calls toDict.

        (ja) このオブジェクトを辞書に変換します。
        戻り値の辞書にはプリミティブ型かプリミティブ型要素のみのリスト
        またはDict等、そしてnullのみを含められます。
        それ以外のクラスを含めたい場合、対象のクラスもこのクラスを継承し、
        toDictを連鎖的に呼び出すようにしてください。
        """
        pass

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "QueryNode":
        """
        (en) Restore this object from the dictionary.

        (ja) このオブジェクトを辞書から復元します。

        Parameters
        ----------
        src: Dict[str, Any]
            A dictionary made with to_dict of this class.
        """
        # 遅延インポート
        from delta_trace_db.query.nodes.logical_node import AndNode, OrNode, NotNode
        from delta_trace_db.query.nodes.comparison_node import FieldEquals, FieldNotEquals, FieldGreaterThan, \
            FieldLessThan, \
            FieldGreaterThanOrEqual, FieldLessThanOrEqual, FieldMatchesRegex, FieldContains, FieldStartsWith, \
            FieldEndsWith, \
            FieldIn, FieldNotIn
        node_type = EnumNodeType[src["type"]]
        match node_type:
            case EnumNodeType.and_:
                return AndNode.from_dict(src)
            case EnumNodeType.or_:
                return OrNode.from_dict(src)
            case EnumNodeType.not_:
                return NotNode.from_dict(src)
            case EnumNodeType.equals_:
                return FieldEquals.from_dict(src)
            case EnumNodeType.notEquals_:
                return FieldNotEquals.from_dict(src)
            case EnumNodeType.greaterThan_:
                return FieldGreaterThan.from_dict(src)
            case EnumNodeType.lessThan_:
                return FieldLessThan.from_dict(src)
            case EnumNodeType.greaterThanOrEqual_:
                return FieldGreaterThanOrEqual.from_dict(src)
            case EnumNodeType.lessThanOrEqual_:
                return FieldLessThanOrEqual.from_dict(src)
            case EnumNodeType.regex_:
                return FieldMatchesRegex.from_dict(src)
            case EnumNodeType.contains_:
                return FieldContains.from_dict(src)
            case EnumNodeType.in_:
                return FieldIn.from_dict(src)
            case EnumNodeType.notIn_:
                return FieldNotIn.from_dict(src)
            case EnumNodeType.startsWith_:
                return FieldStartsWith.from_dict(src)
            case EnumNodeType.endsWith_:
                return FieldEndsWith.from_dict(src)
